#!/usr/bin/env python

__version__ = '$Revision: 4791 $'.split()[1]
__date__ = '$Date: 2006-09-24 14:01:41 -0400 (Sun, 24 Sep 2006) $'.split()[1]
__author__ = 'Kurt Schwehr'

__doc__='''

Tools to generate python code to serialize/deserialize messages
between python and ais binary.  Trying to be as inline as possible, so
no XML on the fly like in ais-py.  

serialize: python to ais binary
deserialize: ais binary to python

The generated code uses translators.py, binary.py, and aisstring.py
which should be packaged with the resulting files.

@requires: U{lxml<http://codespeak.net/lxml/>}
@requires: U{epydoc<http://epydoc.sourceforge.net/>} > 3.0alpha3
@requires: U{BitVector<http://cheeseshop.python.org/pypi/BitVector>}

@author: U{'''+__author__+'''<http://xenon.stanford.edu/~schwehr/>}
@version: ''' + __version__ +'''
@copyright: 2006
@var __date__: Date of last svn commit
@undocumented: __version__ __author__ __doc__ myparser
@since: 2006-Sep-24
@status: under development
@organization: U{CCOM<http://ccom.unh.edu/>}
@license: Restricted while in development to NOAA and USCG.

@todo: add a link to generated doc string to bring up the html for the pretty version

@bug: NOT complete
'''

import sys, os
from decimal import Decimal
from lxml import etree 

def hasSubtag(et,subtag):
    '''
    @return: true if the tag a sub tag with name subtag
    '''
    if 0<len(et.xpath(subtag)): return True
    return False


#def writeBeginning(o,aisBinMsgET):
def writeBeginning(o):
    '''
    Write the doc string header for the message file

    param o: Open output file to write code to.
    param msgET: element tree for the ais message definition.  
    Must be pre-expanded with the expandais.py command.
    '''
    import datetime
    d = datetime.datetime.utcnow()
    dateStr = str(d.year)+'-'+("%02d" %d.month)+'-'+("%02d"%d.day)

    # FIX: what to do for __version__, @since, etc?
    # Need to pass in info about the source file, etc.

    # Minor trickery to get svn to ignore the keywords in the next few lines
    o.write('''#!/usr/bin/env python

__version__ = '$Revision: 4791 $'.split()[1]
__date__ = '$Da'''+'''te: '''+dateStr+''' $'.split()[1]
__author__ = 'xmlbinmsg'

__doc__=\'\'\'

Autogenerated python functions to serialize/deserialize binary messages.

Generated by: '''+__file__+'''

Need to then wrap these functions with the outer AIS packet and then
convert the whole binary blob to a NMEA string.  Those functions are
not currently provided in this file.

serialize: python to ais binary
deserialize: ais binary to python

The generated code uses translators.py, binary.py, and aisstring.py
which should be packaged with the resulting files.

''')

    o.write('''
@requires: U{epydoc<http://epydoc.sourceforge.net/>} > 3.0alpha3
@requires: U{BitVector<http://cheeseshop.python.org/pypi/BitVector>}

@author: \'\'\'+__author__+\'\'\'
@version: \'\'\' + __version__ +\'\'\'
@var __date__: Date of last svn commit
@undocumented: __version__ __author__ __doc__ myparser
@status: under development
@license: Generated code has no license
\'\'\'

import sys
from decimal import Decimal
from BitVector import BitVector

import binary, aisstring

TrueBV  = BitVector(bitstring="1")
"Why always rebuild the True bit?  This should speed things up a bunch"
FalseBV = BitVector(bitstring="0")
"Why always rebuild the False bit?  This should speed things up a bunch"


''')
    return

def generatePython(infile,outfile):
    '''
    @param infile: xml ais binary message definition file
    @param outfile: where to dump the python code
    '''

    aisMsgsET = etree.parse(infile).getroot()

    o = file(outfile,'w')
    os.chmod(outfile,0755)

    writeBeginning(o)

    for msgET in aisMsgsET:
	if msgET.tag != 'message': continue
	print msgET.tag, msgET.attrib['name']

	buildEncode(o,msgET)
	buildDecode(o,msgET)

    print 'FIX: Building unittests... will not actually do anything'
    o.write('\n\n######################################################################\n')
    o.write('# UNIT TESTING\n')
    o.write('######################################################################\n')
    o.write('import unittest\n')

    for msgET in aisMsgsET:
	if msgET.tag != 'message': continue
	print 'Building unit ttests for message ...', msgET.attrib['name']

	buildUnitTest(o,msgET)
	
    buildMain(o)
    return 



######################################################################
# ENCODERS
######################################################################
def encodeBool(o,name,type,numbits,required=None,arraylen=1,unavailable=None, verbose=False):
    '''
    Build the encoder for boolean variables
    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: uint, bool, etc.
    @type numbits: int = 1
    @param numbits: How many bits per unit datum (must be 1 for bools)
    @type required: bool or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many bools will there be?  FIX: handle variable
    @type unavailable: bool or None
    @param unavailable: the default value to use if none given (if not None)
    @return: None
    '''
    
    if verbose: print 'bool encode',name,': unvail=',unavailable

    assert numbits==1
    if arraylen != 1: assert False # FIX... handle arrays
    if verbose: o.write('\t### FIELD: '+name+' (type=bool)\n')
    if None != required:
	assert type(required)==bool
	if required: o.write('\t\tbvList.append(TrueBV)\n')
	else: o.write('\t\tbvList.append(FalseBV)\n')
	if verbose: o.write('\n')
	return

    if None==unavailable:
	o.write('\tif params["'+name+'"]: bvList.append(TrueBV)\n')
	o.write('\telse: bvList.append(FalseBV)\n')
    else: # Have a default value that can be filled in
	assert type(unavailable)==bool
	o.write("\tif '"+name+"' in params:\n")
	o.write('\t\tif params["'+name+'"]: bvList.append(TrueBV)\n')
	o.write('\t\telse: bvList.append(FalseBV)\n')
	o.write('\telse:\n')
	if unavailable: o.write('\t\tbvList.append(TrueBV)\n')
	else: o.write('\t\tbvList.append(FalseBV)\n')
    if verbose: o.write('\n')

def encodeUInt(o,name,type,numbits,required=None,arraylen=1,unavailable=None, verbose=False):
    '''
    Build the encoder for unsigned integer variables

    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: uint, bool, etc.
    @type numbits: int >= 1
    @param numbits: How many bits per unit datum (must be 1..32)
    @type required: bool or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many unsigned ints will there be?  FIX: handle variable
    @type unavailable: bool or None
    @param unavailable: the default value to use if none given (if not None)
    @return: None
    '''
    if verbose: print '  encodeUInt:',name,type,numbits,'Req:',required,'alen:',arraylen,unavailable

    assert numbits>=1 and numbits<=32
    if arraylen != 1: assert False # FIX... handle arrays
    if verbose: o.write('\t### FIELD: '+name+' (type='+type+')\n')

    if None != required:
	if verbose: print '  required:',required
	required=int(required)
	o.write('\tbvList.append(binary.setBitVectorSize(BitVector(intVal='+str(required)+'),'+str(numbits)+'))\n')
	if verbose: o.write('\n')
	return

    if None==unavailable:
	o.write('\tbvList.append(binary.setBitVectorSize(BitVector(intVal=params[\''+name+'\']),'+str(numbits)+'))\n')
    else: # Have a default value that can be filled in
	#assert type(unavailable)==
	int(unavailable) # Make sure unavailable is a number object
	o.write("\tif '"+name+"' in params:\n")
	o.write('\t\tbvList.append(binary.setBitVectorSize(BitVector(intVal=params[\''+name+'\']'+'),'+str(numbits)+'))\n')
	o.write('\telse:\n')
	o.write('\t\tbvList.append(binary.setBitVectorSize(BitVector(intVal='+str(unavailable)+'),'+str(numbits)+'))\n')

    if verbose: o.write('\n')


def encodeFloat(o,name,type,numbits,required=None,arraylen=1,unavailable=None, verbose=False):
    '''
    Build the encoder for IEEE float variables

    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: uint, bool, etc.
    @type numbits: int >= 1
    @param numbits: How many bits per unit datum (must be 1..32)
    @type required: bool or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many unsigned ints will there be?  FIX: handle variable
    @type unavailable: bool or None
    @param unavailable: the default value to use if none given (if not None)
    @return: None
    '''
    if verbose: print '  encodeUInt:',name,type,numbits,'Req:',required,'alen:',arraylen,unavailable

    assert numbits==32 # Force by the IEEE spec
    if arraylen != 1: assert False # FIX... handle arrays
    if verbose: o.write('\t### FIELD: '+name+' (type='+type+')\n')

    if None != required:
	if verbose: print '  required:',required
	required=int(required)
	o.write('\tbvList.append(binary.float2bitvec('+str(required)+'))\n')
	if verbose: o.write('\n')
	return

    if None==unavailable:
	o.write('\tbvList.append(binary.float2bitvec(params[\''+name+'\']))\n')
    else: # Have a default value that can be filled in
	#assert type(unavailable)==
	int(unavailable) # Make sure unavailable is a number object
	o.write("\tif '"+name+"' in params:\n")
	o.write('\t\tbvList.append(binary.float2bitvec(params[\''+name+'\']'+'))\n')
	o.write('\telse:\n')
	o.write('\t\tbvList.append(binary.float2bitvec('+str(unavailable)+'))\n')

    if verbose: o.write('\n')


def encodeAisstr6(o,name,type,numbits,required=None,arraylen=1,unavailable=None, verbose=False):
    '''
    Build the encoder for aisstr6 variables.  Generally are arrays.
    @bug: do we need to optionally check for a valid string?

    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: uint, bool, etc.
    @type numbits: int >= 1
    @param numbits: How many bits per unit datum (must be 1..32)
    @type required: bool or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many unsigned ints will there be?  FIX: handle variable
    @type unavailable: bool or None
    @param unavailable: the default value to use if none given (if not None)
    @return: None
    '''
    if verbose: print '  encodeUInt:',name,type,numbits,'Req:',required,'alen:',arraylen,unavailable

    assert numbits==6 # Force by the IEEE spec
    if verbose: o.write('\t### FIELD: '+name+' (type='+type+')\n')

    if None != required:
	if verbose: print '  required:',required
	required=int(required)
	o.write('\tbvList.append(aisstring.encode(\''+str(required)+'\','+str(numbits*arraylen)+'))\n')
	if verbose: o.write('\n')
	return

    if None==unavailable:
	o.write('\tbvList.append(aisstring.encode(params[\''+name+'\'],'+str(numbits*arraylen)+'))\n')
    else: # Have a default value that can be filled in
	o.write("\tif '"+name+"' in params:\n")
	o.write('\t\tbvList.append(aisstring.encode(params[\''+name+'\'],'+str(numbits*arraylen)+'))\n')
	o.write('\telse:\n')
	o.write('\t\tbvList.append(aisstring.encode(\''+str(unavailable)+'\','+str(numbits*arraylen)+'))\n')

    if verbose: o.write('\n')


def encodeInt(o,name,type,numbits,required=None,arraylen=1,unavailable=None, verbose=False):
    '''
    Build the encoder for signed integer variables

    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: uint, bool, etc.
    @type numbits: int >= 1
    @param numbits: How many bits per unit datum (must be 1..32)
    @type required: bool or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many signed ints will there be?  FIX: handle variable
    @type unavailable: number or None
    @param unavailable: the default value to use if none given (if not None)
    @return: None
    '''
    if verbose: print '  encodeUInt:',name,type,numbits,'Req:',required,'alen:',arraylen,unavailable

    assert numbits>=1 and numbits<=32
    if arraylen != 1: assert False # FIX... handle arrays
    if verbose: o.write('\t### FIELD: '+name+' (type='+type+')\n')

    if None != required:
	if verbose: print '  required:',required
	required=int(required)
	o.write('\tbvList.append(binary.bvFromSignedInt('+str(required)+','+str(numbits)+'))\n')
	if verbose: o.write('\n')
	return


    if None==unavailable:
	o.write('\tbvList.append(binary.bvFromSignedInt(params[\''+name+'\'],'+str(numbits)+'))\n')
    else: # Have a default value that can be filled in
	#assert type(unavailable)==
	int(unavailable) # Make sure unavailable is a number object
	o.write("\tif '"+name+"' in params:\n")
	o.write('\t\tbvList.append(binary.bvFromSignedInt(params[\''+name+'\']'+','+str(numbits)+'))\n')
	o.write('\telse:\n')
	o.write('\t\tbvList.append(binary.bvFromSignedInt('+str(unavailable)+','+str(numbits)+'))\n')

    if verbose: o.write('\n')



# FIX: Ummm... why am I passing the type?  I guess it makes the one print statement easier
def encodeDecimal(o,name,type,numbits,required=None,arraylen=1,unavailable=None, verbose=False, scale=None):
    '''
    Build the encoder for signed decimal variables

    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: decimal
    @type numbits: int >= 1
    @param numbits: How many bits per unit datum (must be 1..32)
    @type required: bool or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many decimals will there be?  FIX: handle variable
    @type unavailable: Decimal or None
    @param unavailable: the default value to use if none given (if not None)
    @return: None
    '''
    if verbose: print '  encodeDecimal:',name,type,numbits,'Req:',required,'alen:',arraylen,unavailable

    assert numbits>=1 and numbits<=32
    if arraylen != 1: assert False # FIX... handle arrays
    if verbose: o.write('\t### FIELD: '+name+' (type='+type+')\n')

    # FIX: optimize to not emit the scaling when it is not needed... or tell the user to switch to an int!
    if None == scale: 
	print 'WARNING: if you are not scaling, then you probably want to use an int instead!'
	print 'Beware canadians bearing travel videos'
	scale='1'

    if None != required:
	if verbose: print '  required:',required
	required=int(required)
	o.write('\tbvList.append(binary.bvFromSignedInt('+str(int(Decimal(required)*Decimal(scale)))+','+str(numbits)+'))\n')
	if verbose: o.write('\n')
	return

    # FIX: can I get rid of the Decimal around params?
    if None==unavailable:
	o.write('\tbvList.append(binary.bvFromSignedInt(int(Decimal(params[\''+name+'\'])*Decimal(\''+scale+'\')),'+str(numbits)+'))\n')
    else: # Have a default value that can be filled in
	o.write("\tif '"+name+"' in params:\n")
	o.write('\t\tbvList.append(binary.bvFromSignedInt(int(Decimal(params[\''+name+'\'])*Decimal(\''+scale+'\')),'+str(numbits)+'))\n')
	o.write('\telse:\n')
	o.write('\t\tbvList.append(binary.bvFromSignedInt('+str(int(Decimal(unavailable)*Decimal(scale)))+','+str(numbits)+'))\n')

    if verbose: o.write('\n')


######################################################################
# DECODERS
######################################################################

def decodeBool(o,name,type,startindex,numbits,required=None,arraylen=1,unavailable=None,
	       bv='bv',dataDict='r',verbose=False):
    '''
    Build the decoder for boolean variables

    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: uint, bool, etc.
    @type startindex: int
    @param startindex: bit that begins the bool(s)
    @type numbits: int = 1
    @param numbits: How many bits per unit datum (must be 1 for bools)
    @type required: bool or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many bools will there be?  FIX: handle variable
    @type unavailable: bool or None
    @param unavailable: the default value to use if none given (if not None)
    @type bv: str
    @param bv: BitVector containing the incoming data
    @type dataDict: str
    @param dataDict: dictionary in which to place the results
    @rtype: int
    @return: index one past the end of where this read
    '''
    assert(type=='bool')
    if verbose: print type,'decode',name,': unvail=',unavailable,'  numbits:',numbits, '  startindex=',startindex
    #int(startindex); int(numbits)  # Make sure it is a number
    assert numbits==1
    assert arraylen == 1 # FIX... handle arrays
    if verbose: o.write('\t### FIELD foo: '+name+' (type='+type+')\n')

    if None != required:
	assert type(required)==bool
	if required: o.write('\t\t'+dataDict+'[\''+name+'\']=True\n')
	else: o.write('\t\t'+dataDict+'[\''+name+'\']=False\n')
	if verbose: o.write('\n')
	return int(startindex)+int(numbits)

    o.write('\t'+dataDict+'[\''+name+'\']=bool(int('+bv+'['+str(startindex)+':'+str(startindex+int(numbits)*int(arraylen))+']))\n')
    if verbose: o.write('\n')

    return int(startindex)+int(numbits)


def decodeUInt(o,name,type,startindex,numbits,required=None,arraylen=1,unavailable=None,
	       bv='bv',dataDict='r',verbose=False):
    '''
    Build the decoder for unsigned integer variables

    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: uint, etc.
    @type startindex: int
    @param startindex: bit that begins the uint(s)
    @type numbits: int >= 1
    @param numbits: How many bits per unit datum
    @type required: int or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many ints will there be?  FIX: handle variable
    @type unavailable: int or None
    @param unavailable: the default value to use if none given (if not None)
    @type bv: str
    @param bv: BitVector containing the incoming data
    @type dataDict: str
    @param dataDict: dictionary in which to place the results
    @rtype: int
    @return: index one past the end of where this read
    '''
    if verbose: print type,'decode',name,': unvail=',unavailable,'  numbits:',numbits, '  startindex=',startindex
    if None==arraylen: arraylen=1
    assert arraylen == 1 # FIX... handle arrays
    assert numbits>=1
    if verbose: o.write('\t### FIELD: '+name+' (type='+type+')\n')

    if None != required:
	int(required) # Make sure required is a number
	o.write('\t'+dataDict+'[\''+name+'\']='+str(required)+'\n')
	if verbose: o.write('\n')
	return startindex+numbits

    o.write('\t'+dataDict+'[\''+name+'\']=int('+bv+'['+str(startindex)+':'+str(startindex+int(numbits)*int(arraylen))+'])\n')
    if verbose: o.write('\n')
    
    return startindex+numbits


def decodeInt(o,name,type,startindex,numbits,required=None,arraylen=1,unavailable=None,
	       bv='bv',dataDict='r',verbose=False):
    '''
    Build the decoder for unsigned integer variables

    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: int
    @type startindex: int
    @param startindex: bit that begins the int(s)
    @type numbits: int >= 1
    @param numbits: How many bits per unit datum
    @type required: int or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many ints will there be?  FIX: handle variable
    @type unavailable: int or None
    @param unavailable: the default value to use if none given (if not None)
    @type bv: str
    @param bv: BitVector containing the incoming data
    @type dataDict: str
    @param dataDict: dictionary in which to place the results
    @rtype: int
    @return: index one past the end of where this read
    '''
    assert type=='int'
    if verbose: print type,'decode',name,': unvail=',unavailable,'  numbits:',numbits, '  startindex=',startindex
    if None==arraylen: arraylen=1
    end = startindex+int(numbits)*int(arraylen)
    assert arraylen == 1 # FIX... handle arrays
    assert numbits>=1
    if verbose: o.write('\t### FIELD: '+name+' (type='+type+')\n')

    if None != required:
	int(required) # Make sure required is a number
	o.write('\t'+dataDict+'[\''+name+'\']='+str(required)+'\n')
	if verbose: o.write('\n')
	return end

    o.write('\t'+dataDict+'[\''+name+'\']=binary.signedIntFromBV('+bv+'['+str(startindex)+':'+str(end)+'])\n')
    if verbose: o.write('\n')
    
    return end

def decodeFloat(o,name,type,startindex,numbits,required=None,arraylen=1,unavailable=None,
	       bv='bv',dataDict='r',verbose=False):
    '''
    Build the decoder for IEEE float variables

    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: int
    @type startindex: int
    @param startindex: bit that begins the int(s)
    @type numbits: int >= 1
    @param numbits: How many bits per unit datum
    @type required: float or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many ints will there be?  FIX: handle variable
    @type unavailable: float or None
    @param unavailable: the default value to use if none given (if not None)
    @type bv: str
    @param bv: BitVector containing the incoming data
    @type dataDict: str
    @param dataDict: dictionary in which to place the results
    @rtype: int
    @return: index one past the end of where this read
    '''
    assert type=='float'
    if verbose: print type,'decode',name,': unvail=',unavailable,'  numbits:',numbits, '  startindex=',startindex
    if None==arraylen: arraylen=1
    end = startindex+int(numbits)*int(arraylen)
    assert arraylen == 1 # FIX... handle arrays
    assert numbits>=1
    if verbose: o.write('\t### FIELD: '+name+' (type='+type+')\n')

    if None != required:
	float(required) # Make sure required is a number
	o.write('\t'+dataDict+'[\''+name+'\']='+str(required)+'\n')
	if verbose: o.write('\n')
	return end

    o.write('\t'+dataDict+'[\''+name+'\']=binary.bitvec2float('+bv+'['+str(startindex)+':'+str(end)+'])\n')
    if verbose: o.write('\n')
    
    return end


def decodeAisstr6(o,name,type,startindex,numbits,required=None,arraylen=1,unavailable=None,
		  bv='bv',dataDict='r',verbose=False):
    '''
    Build the decoder for aisstr6 variables.  Generally arrays.
    @bug: FIX: validate strings??
    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: 'aisstr6'
    @type startindex: int
    @param startindex: bit that begins the int(s)
    @type numbits: int >= 1
    @param numbits: How many bits per unit datum
    @type required: restricted str or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many ints will there be?  FIX: handle variable
    @type unavailable: restricted str or None
    @param unavailable: the default value to use if none given (if not None)
    @type bv: str
    @param bv: BitVector containing the incoming data
    @type dataDict: str
    @param dataDict: dictionary in which to place the results
    @rtype: int
    @return: index one past the end of where this read
    '''
    assert type=='aisstr6'
    if verbose: print type,'decode',name,': unvail=',unavailable,'  numbits:',numbits, '  startindex=',startindex
    if None==arraylen: arraylen=1
    end = startindex+int(numbits)*int(arraylen)
    assert arraylen >= 1 # FIX... handle arrays
    assert numbits>=1
    if verbose: o.write('\t### FIELD: '+name+' (type='+type+')\n')

    if None != required:
	float(required) # Make sure required is a number
	o.write('\t'+dataDict+'[\''+name+'\']='+required+'\n')
	if verbose: o.write('\n')
	return end

    o.write('\t'+dataDict+'[\''+name+'\']=aisstring.decode('+bv+'['+str(startindex)+':'+str(end)+'])\n')
    if verbose: o.write('\n')
    
    return end


def decodeDecimal(o,name,type,startindex,numbits,required=None,arraylen=1,unavailable=None,
	       bv='bv',dataDict='r',verbose=False,scale=None):
    '''
    Build the decoder for signed decimal variables

    @type o: file like obj
    @param o: where write the code 
    @type name: str
    @param name: field name
    @type type: str
    @param type: 'decimal'
    @type startindex: int
    @param startindex: bit that begins the int(s)
    @type numbits: int >= 1
    @param numbits: How many bits per unit datum
    @type required: Decimal or None
    @param required: If not None, then the value must be set to this.
    @type arraylen: int >= 1
    @param arraylen: many ints will there be?  FIX: handle variable
    @type unavailable: Decimal or None
    @param unavailable: the default value to use if none given (if not None)
    @type bv: str
    @param bv: BitVector containing the incoming data
    @type dataDict: str
    @param dataDict: dictionary in which to place the results
    @rtype: int
    @return: index one past the end of where this read
    '''
    assert type=='decimal'
    if verbose: print type,'decode',name,': unvail=',unavailable,'  numbits:',numbits, '  startindex=',startindex
    if None==arraylen: arraylen=1
    end = startindex+int(numbits)*int(arraylen)
    assert arraylen == 1 # FIX... handle arrays
    assert numbits>=1 and numbits <= 32
    if verbose: o.write('\t### FIELD: '+name+' (type='+type+')\n')

    if None == scale: scale='1' # Warning about this was in the encode section

    if None != required:
	Decimal(required) # Make sure required is a number
	o.write('\t'+dataDict+'[\''+name+'\']='+str(Decimal(required))+'/Decimal(\''+scale+'\')\n')
	if verbose: o.write('\n')
	return end

    o.write('\t'+dataDict+'[\''+name+'\']=Decimal(binary.signedIntFromBV('+bv+'['+str(startindex)+':'+str(end)+']))/Decimal(\''+scale+'\')\n')
    if verbose: o.write('\n')
    
    return end



######################################################################
# THE REST
######################################################################



def buildTestParamFunc(o,msgET, verbose=False):
    '''Scrape the testvalues to make a basic param

    @bug: FIX: make this create a dictionary that sits in the overall namespace and spit out deep copies?
    '''
    name = msgET.attrib['name']

    o.write('def '+name+'TestParams():\n')
    o.write("\t'''Return a params file base on the testvalue tags.\n\t@rtype: dict\n\t@return: params based on testvalue tags\n\t'''\n")
    o.write('\tparams = {}\n')
    for field in msgET.xpath('field'):
	name = field.attrib['name']
	type = field.attrib['type']
	if verbose: print 'buildTestParamFunc ...',name,type
	val = None
	if hasSubtag(field,'testvalue') and hasSubtag(field,'required'):
	    print 'ERROR: can not have both test value and required tags in the same field'
	    assert(False)
	if hasSubtag(field,'testvalue'):
	    val = field.xpath('testvalue')[0].text
	else:
	    if not hasSubtag(field,'required'):
		sys.exit("ERROR: missing required or testvalue for field: "+name)
	    val = field.xpath('required')[0].text
	if verbose: print 'buildTestParamFunc for field '+name+' ...',type,val
	o.write('\tparams[\''+name+'\'] = ')
	if type=='bool': 
	    if val=='1' or val.lower=='true': val = 'True'
	    else: val = 'False'
	    o.write(val)
	elif type in ('uint','int','float'):
	    o.write(val)
	elif type in ('decimal','udecimal'):
	    o.write('Decimal(\''+val+'\')')
	elif type in ('aisstr6'):
	    o.write('\''+val+'\'')
	else:
	    print 'ERROR: type not handled ...',type,'  (found in the ',name,' field).  Time to buy more coffee'
	    assert(False)

	o.write('\n')


    o.write('\n\treturn params\n\n')

def buildUnitTest(o,msgET, verbose=False):
    '''
    Write the unittests for a message

    param o: open file where resulting code will be written
    param msgET: Element Tree starting at a message node
    '''
    assert(msgET.tag=='message')
    name = msgET.attrib['name']

    buildTestParamFunc(o,msgET)
    
    o.write('class Test'+name+'(unittest.TestCase):\n')
    o.write("\t'''Uses the testvalue tag text from each type to build a test case for the "+name+" message'''\n")
    o.write('\tdef testEncodeDecode(self):\n')
    # FIX: what is a good docstring to put here?
    o.write('\t\n')
    o.write('\t\tparams = '+name+'TestParams()\n')

    #o.write('\t\tprint "params",params\n')
    o.write('\t\tbits = '+name+'Encode(params)\n')
    #o.write('\t\tprint "bits:",str(bits)\n')
    o.write('\t\tr = '+name+'Decode(bits)\n')
    #o.write('\t\tprint "decoded msg",r\n')
    o.write('\n')
    o.write('\t\t# Check that each parameter came through ok.\n')


    for field in msgET.xpath('field'):
	name = field.attrib['name']
	type = field.attrib['type']
	if type in ('bool','uint','int','aisstr6'):
	    o.write('\t\tself.failUnlessEqual(r[\''+name+'\'],params[\''+name+'\'])\n')
	else:
	    # float, decimal, udecimal
	    # FIX: look up the decimal places if decimal
	    places = '7'
	    if hasSubtag(field,'decimalplaces'): places = field.xpath('decimalplaces')[0].text
	    o.write('\t\tself.failUnlessAlmostEqual(r[\''+name+'\'],params[\''+name+'\'],'+places+')\n')
	


def buildEncode(o,msgET, verbose=False):
    '''
    Write the encoder/decoder for a message

    http://jaynes.colorado.edu/PythonIdioms.html

    param o: open file where resulting code will be written
    param msgET: Element Tree starting at a message node
    '''
    assert(msgET.tag=='message')
    name = msgET.attrib['name']

    print 'Generating encoder for',name # FIX: verbose?
    funcName = name+'Encode'
    o.write('def '+funcName+'(params, validate=False):\n')

    ########################################
    # doc string
    o.write("\t'''Create a "+name+" binary message payload to pack into an AIS Msg 8 (binary message).\n\n")
    o.write('\tFields in params:\n')
    for field in msgET.xpath('field'):
	desc = field[0].text.replace('\n',' ') # get ride of new lines
	o.write('\t  - '+field.attrib['name']+'('+field.attrib['type']+'): '+desc) # give the description
	if len(field.xpath("required")) == 1:
	    o.write(' (field automatically set to "'+field.xpath("required")[0].text+'")')

	o.write('\n')
    o.write('\t@param params: Dictionary of field names/values.  Throws a ValueError exception if required is missing\n')
    o.write('\t@param validate: Set to true to cause checking to occur.  Runs slower.  FIX: not implemented.\n')
	
    o.write("\t@rtype: BitVector\n")
    o.write("\t@return: encoded binary message ready for wrapping in an AIS Msg 8\n")
    o.write("\t'''\n\n")

    ########################################
    # Actually build the code

    o.write('\tbvList = []\n')

    if verbose: print 'number of fields = ', len(msgET.xpath('field'))

    dynamicArrays = False # Set to true when positioning must be calculated

    for field in msgET.xpath('field'):
	name = field.attrib['name']
	type = field.attrib['type']
        numbits = int(field.attrib['numberofbits'])
	required = None; 
	if hasSubtag(field,'required'): 
	    required = field.xpath('required')[0].text
	    #print 'required set for',name,'to',required
	unavailable=None; 
	if hasSubtag(field,'unavailable'): unavailable = field.xpath('unavailable')[0].text
	arraylen=1
	if 'arraylength' in field.attrib: 
	    arraylen=int(field.attrib['arraylength'])
	    if verbose: print 'Processing field ...',name,'('+type+' ',numbits,'*',arraylen,'=',numbits*arraylen,'bits )'
	else:
	    if verbose: print 'Processing field ...',name,'(',type+' ',numbits,')'

	if   type=='bool'   : encodeBool   (o,name,type,numbits,required,arraylen,unavailable)
	elif type=='uint'   : encodeUInt   (o,name,type,numbits,required,arraylen,unavailable)
	elif type=='int'    : encodeInt    (o,name,type,numbits,required,arraylen,unavailable)	    
	elif type=='float'  : encodeFloat  (o,name,type,numbits,required,arraylen,unavailable)
	elif type=='aisstr6': encodeAisstr6(o,name,type,numbits,required,arraylen,unavailable)	    
        elif type=='decimal':
	    scale = None
	    if hasSubtag(field,'scale'): scale = field.xpath('scale')[0].text
	    encodeDecimal(o,name,type,numbits,required,arraylen,unavailable,scale=scale)
        else: print 'WARNING: In buildEncode - Unhandled field type for',name,'...',type

    o.write('\n\treturn binary.joinBV(bvList)\n\n')


######################################################################
# DECODER RING

def buildDecode(o,msgET, verbose=False):
    '''
    Write the decoder for a message

    param o: open file where resulting code will be written
    type msgET: elementtree
    param msgET: Element Tree starting at a message node
    return: None
    '''
    assert(msgET.tag=='message')
    name = msgET.attrib['name']

    print 'Generating decoder for',name
    funcName = name+'Decode'
    o.write('def '+funcName+'(bv, validate=False):\n')

    ########################################
    # doc string
    o.write("\t'''Unpack a "+name+" binary message bit payload from an AIS Msg 8 (binary message)\n\n")
    o.write('\tFields in params:\n')
    for field in msgET.xpath('field'):
	desc = field[0].text.replace('\n',' ') # get ride of new lines
	o.write('\t  - '+field.attrib['name']+'('+field.attrib['type']+'): '+desc) # give the description
	if len(field.xpath("required")) == 1:
	    o.write(' (field automatically set to "'+field.xpath("required")[0].text+'")')

	o.write('\n')
    o.write('\t@type bv: BitVector\n')
    o.write('\t@param bv: Bits defining a message\n')
    o.write('\t@param validate: Set to true to cause checking to occur.  Runs slower.  FIX: not implemented.\n')
	
    o.write("\t@rtype: dict\n")
    o.write("\t@return: params\n")
    o.write("\t'''\n\n")

    ########################################
    # Actually build the code

    o.write('\tr = {}\n')

    if verbose: print 'number of fields = ', len(msgET.xpath('field'))

    dynamicArrays = False # Set to true when positioning must be calculated

    startindex = 0 # Where we are in the bitvector... FIX: what about variable length jobs?

    for field in msgET.xpath('field'):
	name = field.attrib['name']
	type = field.attrib['type']
        numbits = int(field.attrib['numberofbits'])
	required = None; 
	if hasSubtag(field,'required'): 
	    required = field.xpath('required')[0].text
	    print 'required set for',name,'to',required
	unavailable=None; 
	if hasSubtag(field,'unavailable'): unavailable = field.xpath('unavailable')[0].text
	arraylen=1
	if 'arraylength' in field.attrib: 
	    arraylen=int(field.attrib['arraylength'])
	    if verbose: print 'Processing field ...',name,'('+type+' ',numbits,'*',arraylen,'=',numbits*arraylen,'bits )'
	else:
	    if verbose: print 'Processing field ...',name,'(',type+' ',numbits,')'

	assert None!=startindex
	if verbose: print 'startindex',startindex
	if   type=='bool'   : startindex = decodeBool   (o,name,type,startindex,numbits,required,arraylen,unavailable)
	elif type=='uint'   : startindex = decodeUInt   (o,name,type,startindex,numbits,required,arraylen,unavailable)
	elif type=='int'    : startindex = decodeInt    (o,name,type,startindex,numbits,required,arraylen,unavailable)
	elif type=='float'  : startindex = decodeFloat  (o,name,type,startindex,numbits,required,arraylen,unavailable)
	elif type=='aisstr6': startindex = decodeAisstr6(o,name,type,startindex,numbits,required,arraylen,unavailable)	    
        elif type=='decimal':
	    scale = None
	    if hasSubtag(field,'scale'): scale = field.xpath('scale')[0].text
	    #print 'pre call...'
	    #print '  required: "'+str(required)+'"  scale: "'+str(scale)+'"  unavail:', unavailable
	    startindex = decodeDecimal(o,name,type,startindex, numbits,required,arraylen,unavailable,scale=scale)
        else: print 'WARNING: In buildDecode - Unhandled field type for',name,'...',type

	if None==startindex: print 'FIX: here. drat.  treat me right'
	assert None!=startindex


    o.write('\treturn r\n')




def buildMain(o):
    o.write('''
############################################################
if __name__=='__main__':

    from optparse import OptionParser
    myparser = OptionParser(usage="%prog [options]",
			    version="%prog "+__version__)

    #sys.exit('EARLY EXIT - FIX: remove')
    myparser.add_option('--doc-test',dest='doctest',default=False,action='store_true',
                        help='run the documentation tests')
    myparser.add_option('--unit-test',dest='unittest',default=False,action='store_true',
                        help='run the unit tests')
    myparser.add_option('-v','--verbose',dest='verbose',default=False,action='store_true',
                        help='Make the test output verbose')

    (options,args) = myparser.parse_args()
    success=True

    if options.doctest:
	import os; print os.path.basename(sys.argv[0]), 'doctests ...',
	sys.argv= [sys.argv[0]]
	if options.verbose: sys.argv.append('-v')
	import doctest
	numfail,numtests=doctest.testmod()
	if numfail==0: print 'ok'
	else: 
	    print 'FAILED'
	    success=False

    if not success:
	sys.exit('Something Failed')

    del success # Hide success from epydoc

    if options.unittest:
	sys.argv = [sys.argv[0]]
	if options.verbose: sys.argv.append('-v')
	unittest.main()
''')

######################################################################
if __name__=='__main__':
    from optparse import OptionParser
    myparser = OptionParser(usage="%prog [options]",
			    version="%prog "+__version__)

    myparser.add_option('-o','--output',dest='outputFileName',default=None,
			help='Name of the python file to write')
#			help='Name of the python file to write [default: stdout]')

    myparser.add_option('-i','-x','--xml-definition',dest='xmlFileName',default=None,
			help='XML definition file for the msg to use')
#			help='XML definition file for the msg to use [default: stdin]')

#    myparser.add_option('-m','--message',dest='message',default=None,
#			help='Which message to write code for [default: all]')


    (options,args) = myparser.parse_args()

    #infile=sys.stdin
    #if options.xmlFileName: infile = file(options.xmlFileName,'r')

    #outfile=sys.stdout
    #if options.outputFile: outfile = file(options.xmlDefinition,'w')

    # FIX: down the road, it would be good to allow either filenames of std{in/out}

    if None==options.xmlFileName:
	sys.exit('ERROR: must specify an xml definition file.')
    if None==options.outputFileName:
	sys.exit('ERROR: must specify an python file to write to.')
    generatePython(options.xmlFileName,options.outputFileName)

    print '\nrecommend running pychecker like this:'
    print '  pychecker -q',options.outputFileName
